@@ -2,8 +2,6 @@
 # License, v. 2.0. If a copy of the MPL was not distributed with this
 # file, You can obtain one at http://mozilla.org/MPL/2.0/.
 
-
-import json
 import logging
 import os
 import shutil
@@ -11,6 +9,12 @@ import sys
 import time
 from collections import defaultdict
 
+try:
+    import orjson
+except ImportError:
+    orjson = None
+    import json
+
 import yaml
 from redo import retry
 from taskgraph import create
@@ -35,10 +39,10 @@ from .actions import render_actions_json
 from .files_changed import get_changed_files
 from .parameters import get_app_version, get_version
 from .try_option_syntax import parse_message
-from .util.backstop import BACKSTOP_INDEX, is_backstop
+from .util.backstop import ANDROID_PERFTEST_BACKSTOP_INDEX, BACKSTOP_INDEX, is_backstop
 from .util.bugbug import push_schedules
 from .util.chunking import resolver
-from .util.hg import get_hg_commit_message, get_hg_revision_branch
+from .util.hg import get_hg_commit_message, get_hg_revision_branch, get_hg_revision_info
 from .util.partials import populate_release_history
 from .util.taskcluster import insert_index
 from .util.taskgraph import find_decision_task, find_existing_tasks_from_previous_kinds
@@ -92,9 +96,9 @@ PER_PROJECT_PARAMETERS = {
         "target_tasks_method": "mozilla_release_tasks",
         "release_type": "release",
     },
-    "mozilla-esr128": {
-        "target_tasks_method": "mozilla_esr128_tasks",
-        "release_type": "esr128",
+    "mozilla-esr140": {
+        "target_tasks_method": "mozilla_esr140_tasks",
+        "release_type": "esr140",
     },
     "pine": {
         "target_tasks_method": "pine_tasks",
@@ -120,6 +124,20 @@ PER_PROJECT_PARAMETERS = {
 }
 
 
+def load_json(fh):
+    if orjson is not None:
+        return orjson.loads(fh.read())
+    else:
+        return json.load(fh)
+
+
+def dump_json(fh, data):
+    if orjson is not None:
+        fh.write(orjson.dumps(data))
+    else:
+        fh.write(json.dumps(data, separators=(",", ": ")).encode("utf-8"))
+
+
 def full_task_graph_to_runnable_jobs(full_task_json):
     runnable_jobs = {}
     for label, node in full_task_json.items():
@@ -231,12 +249,27 @@ def taskgraph_decision(options, parameters=None):
     if len(push_schedules) > 0:
         write_artifact("bugbug-push-schedules.json", push_schedules.popitem()[1])
 
-    # cache run-task & misc/fetch-content
+    # cache run-task, misc/fetch-content & robustcheckout.py
     scripts_root_dir = os.path.join(GECKO, "taskcluster/scripts")
     run_task_file_path = os.path.join(scripts_root_dir, "run-task")
-    fetch_content_file_path = os.path.join(scripts_root_dir, "misc/fetch-content")
+    test_linux_file_path = os.path.join(scripts_root_dir, "tester", "test-linux.sh")
+    fetch_content_file_path = os.path.join(
+        GECKO,
+        "third_party",
+        "python",
+        "taskcluster_taskgraph",
+        "taskgraph",
+        "run-task",
+        "fetch-content",
+    )
+    robustcheckout_path = os.path.join(
+        GECKO,
+        "testing/mozharness/external_tools/robustcheckout.py",
+    )
     shutil.copy2(run_task_file_path, ARTIFACTS_DIR)
+    shutil.copy2(test_linux_file_path, ARTIFACTS_DIR)
     shutil.copy2(fetch_content_file_path, ARTIFACTS_DIR)
+    shutil.copy2(robustcheckout_path, ARTIFACTS_DIR)
 
     # actually create the graph
     create_tasks(
@@ -297,6 +330,11 @@ def get_decision_parameters(graph_config, options):
         env_prefix=_get_env_prefix(graph_config),
     )
 
+    if head_git_rev := get_hg_revision_info(
+        GECKO, revision=parameters["head_rev"], info="extras.git_commit"
+    ):
+        parameters["head_git_rev"] = head_git_rev
+
     # Define default filter list, as most configurations shouldn't need
     # custom filters.
     parameters["filters"] = [
@@ -352,9 +390,9 @@ def get_decision_parameters(graph_config, options):
         parameters.update(PER_PROJECT_PARAMETERS[project])
     except KeyError:
         logger.warning(
-            "using default project parameters; add {} to "
-            "PER_PROJECT_PARAMETERS in {} to customize behavior "
-            "for this project".format(project, __file__)
+            f"using default project parameters; add {project} to "
+            f"PER_PROJECT_PARAMETERS in {__file__} to customize behavior "
+            "for this project"
         )
         parameters.update(PER_PROJECT_PARAMETERS["default"])
 
@@ -394,6 +432,14 @@ def get_decision_parameters(graph_config, options):
     # Determine if this should be a backstop push.
     parameters["backstop"] = is_backstop(parameters)
 
+    # For the android perf tasks, run them 50% less often
+    parameters["android_perftest_backstop"] = is_backstop(
+        parameters,
+        push_interval=30,
+        time_interval=60 * 6,
+        backstop_strategy="android_perftest_backstop",
+    )
+
     if "decision-parameters" in graph_config["taskgraph"]:
         find_object(graph_config["taskgraph"]["decision-parameters"])(
             graph_config, parameters
@@ -431,8 +477,8 @@ def get_existing_tasks(rebuild_kinds, parameters, graph_config):
 def set_try_config(parameters, task_config_file):
     if os.path.isfile(task_config_file):
         logger.info(f"using try tasks from {task_config_file}")
-        with open(task_config_file) as fh:
-            task_config = json.load(fh)
+        with open(task_config_file, "rb") as fh:
+            task_config = load_json(fh)
         task_config_version = task_config.pop("version", 1)
         if task_config_version == 1:
             parameters["try_mode"] = "try_task_config"
@@ -454,6 +500,8 @@ def set_try_config(parameters, task_config_file):
 
 def set_decision_indexes(decision_task_id, params, graph_config):
     index_paths = []
+    if params["android_perftest_backstop"]:
+        index_paths.insert(0, ANDROID_PERFTEST_BACKSTOP_INDEX)
     if params["backstop"]:
         # When two Decision tasks run at nearly the same time, it's possible
         # they both end up being backstops if the second checks the backstop
@@ -477,13 +525,13 @@ def write_artifact(filename, data):
         with open(path, "w") as f:
             yaml.safe_dump(data, f, allow_unicode=True, default_flow_style=False)
     elif filename.endswith(".json"):
-        with open(path, "w") as f:
-            json.dump(data, f, sort_keys=True, indent=2, separators=(",", ": "))
+        with open(path, "wb") as f:
+            dump_json(f, data)
     elif filename.endswith(".json.gz"):
         import gzip
 
         with gzip.open(path, "wb") as f:
-            f.write(json.dumps(data).encode("utf-8"))
+            dump_json(f, data)
     else:
         raise TypeError(f"Don't know how to write to {filename}")
 
@@ -493,13 +541,13 @@ def read_artifact(filename):
     if filename.endswith(".yml"):
         return load_yaml(path, filename)
     if filename.endswith(".json"):
-        with open(path) as f:
-            return json.load(f)
+        with open(path, "rb") as f:
+            return load_json(f)
     if filename.endswith(".json.gz"):
         import gzip
 
         with gzip.open(path, "rb") as f:
-            return json.load(f.decode("utf-8"))
+            return load_json(f)
     else:
         raise TypeError(f"Don't know how to read {filename}")
 
